import pytest

from pathlib import Path
import os.path as osp
import importlib
from copy import copy
import pickle

import numpy as np
from scipy.spatial.distance import euclidean

import simtk.openmm.app as omma
import simtk.openmm as omm
import simtk.unit as unit

import mdtraj as mdj

from openmm_systems.test_systems import LennardJonesPair

from wepy.sim_manager import Manager

### Apparatus

## Resampler
from wepy.resampling.distances.distance import Distance
from wepy.resampling.resamplers.wexplore import WExploreResampler
from wepy.resampling.resamplers.revo import REVOResampler
from wepy.resampling.resamplers.resampler import NoResampler

## Boundary Conditions
from wepy.boundary_conditions.unbinding import UnbindingBC

## Runner
from wepy.runners.openmm import (
    OpenMMRunner,
    OpenMMState,
    OpenMMWalker,
    UNIT_NAMES,
    GET_STATE_KWARG_DEFAULTS,
    gen_walker_state,
)

## Initial Walkers
from wepy.walker import Walker

### Configuration

## Reporters
from wepy.reporter.hdf5 import WepyHDF5Reporter
from wepy.reporter.restree import ResTreeReporter
from wepy.reporter.dashboard import DashboardReporter

## Work Mappers
from wepy.work_mapper.mapper import Mapper
from wepy.work_mapper.mapper import WorkerMapper
from wepy.work_mapper.task_mapper import TaskMapper

### Utilities
from wepy.util.mdtraj import mdtraj_to_json_topology

### Orchestration
from wepy.orchestration.configuration import Configuration
from wepy.orchestration.orchestrator import Orchestrator, reconcile_orchestrators
from wepy.orchestration.snapshot import WepySimApparatus, SimSnapshot

### Mock Systems

from wepy_tools.sim_makers.openmm.sim_maker import OpenMMSimMaker
from wepy_tools.systems.lennard_jones import PairDistance


### Constants

# only use the reference platform for python-only integration testing
# purposes
PLATFORM = "Reference"

### Sanity Test
@pytest.fixture(scope='class')
def lj_sanity_test():
    """Sanity test to make sure we even have the plugin fixtures installed."""
    return "sanity"

### Fixtures


## OpenMM Misc.

@pytest.fixture(scope='class')
def lj_omm_sys():
    return LennardJonesPair()

@pytest.fixture(scope='class')
def langevin_integrator():

    integrator = omm.LangevinIntegrator(
        *OpenMMSimMaker.DEFAULT_INTEGRATOR_PARAMS['LangevinIntegrator']
    )

    return integrator

integrators = [
    langevin_integrator,
]

@pytest.fixture(
    scope='class',
    params=[
        'LangevinIntegrator',
    ]
)
def lj_integrator(
        request,
        *integrators,
):

    intgr_spec = request.param
    if intgr_spec == "LangevinIntegrator":
        return langevin_integrator
    else:
        raise ValueError("Unkown integrator")


## Runner

@pytest.fixture(
    scope='class',
    params=[
        'Reference',
    ]
)
def lj_openmm_runner(
        request,
        lj_omm_sys,
        lj_integrator
):

    # parametrize the platform
    platform = request.param

    positions = test_sys.positions.value_in_unit(test_sys.positions.unit)

    init_state = gen_walker_state(
        positions,
        test_sys.system,
        integrator)

    # initialize the runner
    runner = OpenMMRunner(
        lj_omm_sys.system,
        lj_omm_sys.topology,
        lj_integrator,
        platform=platform)

    return runner

## Resampler

# @pytest.fixture(scope='class')
# def lj_distance_metric():
#     return PairDistance()


# @pytest.fixture(scope='class')
# def lj_wexplore_resampler(lj_params, lj_distance_metric, lj_init_state):
#     resampler = WExploreResampler(distance=lj_distance_metric,
#                                    init_state=lj_init_state,
#                                    max_region_sizes=lj_params['max_region_sizes'],
#                                    max_n_regions=lj_params['max_n_regions'],
#                                    pmin=lj_params['pmin'], pmax=lj_params['pmax'])

#     return resampler

# @pytest.fixture(scope='class')
# def lj_revo_resampler(lj_params, lj_distance_metric, lj_init_state):
#     resampler = REVOResampler(distance=lj_distance_metric,
#                               merge_dist=2.5,
#                               char_dist=1.0,
#                               init_state=lj_init_state,
#                               pmin=lj_params['pmin'], pmax=lj_params['pmax'])

#     return resampler

# @pytest.fixture(scope='class')
# def lj_topology(lj_omm_sys):

#     # the mdtraj here is needed for the distance function
#     mdtraj_topology = mdj.Topology.from_openmm(lj_omm_sys.topology)

#     ## Reporters if we want them
#     json_str_top = mdtraj_to_json_topology(mdtraj_topology)

#     return json_str_top


# @pytest.fixture(scope='class')
# def lj_unbinding_bc(lj_params, lj_init_state, lj_topology, lj_omm_sys):

#     # initialize the unbinding boundary condition
#     ubc = UnbindingBC(cutoff_distance=lj_params['cutoff_distance'],
#                       initial_state=lj_init_state,
#                       topology=lj_topology,
#                       ligand_idxs=np.array(lj_omm_sys.ligand_indices),
#                       receptor_idxs=np.array(lj_omm_sys.receptor_indices))

#     return ubc

# @pytest.fixture(scope='class')
# def lj_reporter_kwargs(lj_params, lj_topology, lj_wexplore_resampler, lj_unbinding_bc):
#     """Reporters that work for all of the components."""

#     # make a dictionary of units for adding to the HDF5
#     units = dict(UNIT_NAMES)

#     hdf5_reporter_kwargs = {'save_fields' : lj_params['save_fields'],
#                             'resampler' : lj_wexplore_resampler,
#                             'boundary_conditions' : lj_unbinding_bc,
#                             'topology' : lj_topology,
#                             'units' : units,
#     }

#     dashboard_reporter_kwargs = {'step_time' : lj_params['step_size'].value_in_unit(unit.second),
#                                  'bc_cutoff_distance' : lj_unbinding_bc.cutoff_distance}

#     # Resampling Tree
#     restree_reporter_kwargs = {'resampler' : lj_wexplore_resampler,
#                                'boundary_condition' : lj_unbinding_bc,
#                                'node_radius' : 3.0,
#                                'row_spacing' : 5.0,
#                                'step_spacing' : 20.0,
#                                'progress_key' : 'min_distances',
#                                'max_progress_value' : lj_unbinding_bc.cutoff_distance,
#                                'colormap_name' : 'plasma'}


#     reporter_kwargs = [hdf5_reporter_kwargs, dashboard_reporter_kwargs,
#                        restree_reporter_kwargs]

#     return reporter_kwargs


# @pytest.fixture(scope='class')
# def lj_reporter_classes():
#     reporter_classes = [WepyHDF5Reporter, DashboardReporter,
#                         ResTreeReporter]

#     return reporter_classes

# @pytest.fixture(scope='class')
# def lj_init_walkers(lj_params, lj_init_sim_state):
#     init_weight = 1.0 / lj_params['n_walkers']
#     init_walkers = [OpenMMWalker(OpenMMState(lj_init_sim_state), init_weight)
#                     for i in range(lj_params['n_walkers'])]

#     return init_walkers


# @pytest.fixture(scope='class')
# def lj_apparatus(lj_openmm_runner, lj_wexplore_resampler, lj_unbinding_bc):

#     sim_apparatus = WepySimApparatus(lj_openmm_runner, resampler=lj_wexplore_resampler,
#                                      boundary_conditions=lj_unbinding_bc)

#     return sim_apparatus

# @pytest.fixture(scope='class')
# def lj_null_apparatus(lj_openmm_runner):

#     sim_apparatus = WepySimApparatus(lj_openmm_runner, resampler=NoResampler())

#     return sim_apparatus

# @pytest.fixture(scope='class')
# def lj_snapshot(lj_init_walkers, lj_apparatus):

#     return SimSnapshot(lj_init_walkers, lj_apparatus)



# @pytest.fixture(scope='class')
# def lj_configuration(tmp_path_factory, lj_reporter_classes, lj_reporter_kwargs):

#     # make a temporary directory for this configuration to work with
#     tmpdir = str(tmp_path_factory.mktemp('lj_fixture'))
#     # tmpdir = tmp_path_factory.mktemp('lj_fixture/work_dir')

#     configuration = Configuration(work_dir=tmpdir,
#                                   reporter_classes=lj_reporter_classes,
#                                   reporter_partial_kwargs=lj_reporter_kwargs)

#     return configuration

# @pytest.fixture(scope='class')
# def lj_null_configuration(tmp_path_factory, lj_reporter_classes, lj_reporter_kwargs,
#                           lj_params, lj_wexplore_resampler, lj_topology):

#     reporter_classes = [WepyHDF5Reporter]

#     # make a dictionary of units for adding to the HDF5
#     units = dict(UNIT_NAMES)

#     hdf5_reporter_kwargs = {'save_fields' : lj_params['save_fields'],
#                             'resampler' : lj_wexplore_resampler,
#                             'topology' : lj_topology,
#                             'units' : units,
#     }

#     reporter_kwargs = [hdf5_reporter_kwargs]



#     # make a temporary directory for this configuration to work with
#     tmpdir = str(tmp_path_factory.mktemp('lj_fixture'))
#     # tmpdir = tmp_path_factory.mktemp('lj_fixture/work_dir')

#     configuration = Configuration(work_dir=tmpdir,
#                                   reporter_classes=reporter_classes,
#                                   reporter_partial_kwargs=reporter_kwargs)

#     return configuration

# @pytest.fixture(scope='class')
# def lj_inmem_configuration(tmp_path_factory):

#     # make a temporary directory for this configuration to work with
#     tmpdir = str(tmp_path_factory.mktemp('lj_fixture'))
#     # tmpdir = tmp_path_factory.mktemp('lj_fixture/work_dir')

#     configuration = Configuration(work_dir=tmpdir)

#     return configuration


# @pytest.fixture(scope='class')
# def lj_work_mapper(lj_configuration):

#     work_mapper = Mapper()

#     return work_mapper


# @pytest.fixture(scope='class')
# def lj_work_mapper_worker():

#     work_mapper = WorkerMapper(num_workers=1)

#     return work_mapper

# @pytest.fixture(scope='class')
# def lj_work_mapper_task():

#     work_mapper = TaskMapper(num_workers=1)

#     return work_mapper


# @pytest.fixture(scope='class')
# def lj_reporters(tmp_path_factory, lj_reporter_classes, lj_reporter_kwargs):


#     # make a temporary directory for this configuration to work with
#     tmpdir = str(tmp_path_factory.mktemp('lj_fixture'))

#     # make a config so that the reporters get parametrized properly
#     config = Configuration(work_dir=tmpdir,
#                            reporter_classes=lj_reporter_classes,
#                            reporter_partial_kwargs=lj_reporter_kwargs)

#     return config.reporters




# @pytest.fixture(scope='class')
# def lj_orchestrator(lj_apparatus, lj_init_walkers, lj_configuration):

#     # use an in memory database with sqlite

#     # make a path to the temporary directory for this orchestrator
#     # orch_path = tmp_path_factory.mktemp('lj_fixture/lj.orch.sqlite')

#     # then create the seed/root/master orchestrator which will be used
#     # from here on out
#     orch = Orchestrator()

#     return orch

# @pytest.fixture(scope='class')
# def lj_orchestrator_defaults(lj_orchestrator,
#                                  lj_apparatus, lj_init_walkers, lj_configuration):


#     lj_orchestrator.set_default_sim_apparatus(lj_apparatus)
#     lj_orchestrator.set_default_init_walkers(lj_init_walkers)
#     lj_orchestrator.set_default_configuration(lj_configuration)

#     lj_orchestrator.gen_default_snapshot()


#     return lj_orchestrator

# @pytest.fixture(scope='class')
# def lj_orchestrator_defaults_inmem(lj_orchestrator,
#                                        lj_apparatus, lj_init_walkers, lj_inmem_configuration):


#     lj_orchestrator.set_default_sim_apparatus(lj_apparatus)
#     lj_orchestrator.set_default_init_walkers(lj_init_walkers)
#     lj_orchestrator.set_default_configuration(lj_configuration)

#     lj_orchestrator.gen_default_snapshot()


#     return lj_orchestrator

# @pytest.fixture(scope='class')
# def lj_orchestrator_defaults_null(lj_orchestrator,
#                                   lj_null_apparatus, lj_init_walkers,
#                                   lj_null_configuration):

#     lj_orchestrator.set_default_sim_apparatus(lj_null_apparatus)
#     lj_orchestrator.set_default_init_walkers(lj_init_walkers)
#     lj_orchestrator.set_default_configuration(lj_null_configuration)

#     lj_orchestrator.gen_default_snapshot()


#     return lj_orchestrator



# @pytest.fixture(scope='class')
# def lj_orchestrator_file(tmp_path_factory, lj_apparatus, lj_init_walkers, lj_configuration):

#     # use an in memory database with sqlite

#     # make a path to the temporary directory for this orchestrator
#     orch_path = str(tmp_path_factory.mktemp('lj_fixture') / "lj.orch.sqlite")

#     # then create the seed/root/master orchestrator which will be used
#     # from here on out
#     orch = Orchestrator(orch_path)

#     return orch

# @pytest.fixture(scope='class')
# def lj_orchestrator_file_other(tmp_path_factory,
#                                    lj_apparatus, lj_init_walkers, lj_configuration):

#     # use an in memory database with sqlite

#     # make a path to the temporary directory for this orchestrator
#     orch_path = str(tmp_path_factory.mktemp('lj_fixture') / "lj_other.orch.sqlite")

#     # then create the seed/root/master orchestrator which will be used
#     # from here on out
#     orch = Orchestrator(orch_path)

#     return orch

# @pytest.fixture(scope='class')
# def lj_orchestrator_defaults_file(lj_orchestrator_file,
#                                       lj_apparatus, lj_init_walkers, lj_configuration):


#     lj_orchestrator_file.set_default_sim_apparatus(lj_apparatus)
#     lj_orchestrator_file.set_default_init_walkers(lj_init_walkers)
#     lj_orchestrator_file.set_default_configuration(lj_configuration)

#     lj_orchestrator_file.gen_default_snapshot()


#     return lj_orchestrator_file

# @pytest.fixture(scope='class')
# def lj_orchestrator_defaults_file_other(lj_orchestrator_file_other,
#                                             lj_apparatus, lj_init_walkers, lj_configuration):

#     lj_orchestrator_file_other.set_default_sim_apparatus(lj_apparatus)
#     lj_orchestrator_file_other.set_default_init_walkers(lj_init_walkers)
#     lj_orchestrator_file_other.set_default_configuration(lj_configuration)

#     lj_orchestrator_file_other.gen_default_snapshot()


#     return lj_orchestrator_file_other


# @pytest.fixture(scope='class')
# def lj_sim_manager(tmp_path_factory, lj_orchestrator_defaults):

#     start_snapshot = lj_orchestrator_defaults.get_default_snapshot()
#     configuration = lj_orchestrator_defaults.get_default_configuration()

#     # make a new temp dir for this configuration
#     tempdir = str(tmp_path_factory.mktemp('lj_sim_manager'))
#     configuration = configuration.reparametrize(work_dir=tempdir)

#     sim_manager = lj_orchestrator_defaults.gen_sim_manager(start_snapshot,
#                                                                configuration=configuration)

#     return sim_manager


# @pytest.fixture(scope='class')
# def lj_sim_manager_inmem(tmp_path_factory, lj_orchestrator_defaults_inmem):

#     start_snapshot = lj_orchestrator_defaults_inmem.get_default_snapshot()
#     configuration = lj_orchestrator_defaults_inmem.get_default_configuration()

#     # make a new temp dir for this configuration
#     tempdir = str(tmp_path_factory.mktemp('lj_sim_manager'))
#     configuration = configuration.reparametrize(work_dir=tempdir)

#     sim_manager = lj_orchestrator_defaults.gen_sim_manager(start_snapshot,
#                                                            configuration=configuration)

#     return sim_manager

# @pytest.fixture(scope='class')
# def lj_sim_manager_null(tmp_path_factory, lj_orchestrator_defaults_null):

#     start_snapshot = lj_orchestrator_defaults_null.get_default_snapshot()
#     configuration = lj_orchestrator_defaults_null.get_default_configuration()

#     # make a new temp dir for this configuration
#     tempdir = str(tmp_path_factory.mktemp('lj_sim_manager'))
#     configuration = configuration.reparametrize(work_dir=tempdir)

#     sim_manager = lj_orchestrator_defaults_null.gen_sim_manager(start_snapshot,
#                                                                 configuration=configuration)

#     return sim_manager


# @pytest.fixture(scope='class')
# def lj_sim_manager_run_results(lj_sim_manager):

#     n_cycles = 10
#     n_steps = 100

#     steps = [n_steps for _ in range(n_cycles)]

#     return lj_sim_manager.run_simulation(n_cycles, steps)

# @pytest.fixture(scope='class')
# def lj_sim_manager_null_run_results(lj_sim_manager_null):

#     n_cycles = 10
#     n_steps = 100

#     steps = [n_steps for _ in range(n_cycles)]

#     return lj_sim_manager_null.run_simulation(n_cycles, steps)

# @pytest.fixture(scope='class')
# def lj_orch_run_by_time_results(tmp_path_factory, lj_orchestrator_defaults):

#     runtime = 20 # seconds
#     n_steps = 100

#     start_snaphash = lj_orchestrator_defaults.get_default_snapshot_hash()

#     # make a new temp dir for this configuration
#     configuration = lj_orchestrator_defaults.get_default_configuration()
#     tempdir = str(tmp_path_factory.mktemp('lj_sim_manager'))
#     configuration = configuration.reparametrize(work_dir=tempdir)


#     return lj_orchestrator_defaults.run_snapshot_by_time(start_snaphash,
#                                                              runtime, n_steps,
#                                                              configuration=configuration)

# @pytest.fixture(scope='class')
# def lj_orch_run_end_snapshot(lj_orch_run_by_time_results):

#     end_snapshot, _, _, _ = lj_orch_run_by_time_results

#     return end_snapshot

# @pytest.fixture(scope='class')
# def lj_orch_orchestrated_run(tmp_path_factory, lj_orchestrator_defaults):

#     run_time = 20 # seconds
#     n_steps = 100

#     start_snaphash = lj_orchestrator_defaults.get_default_snapshot_hash()

#     tempdir = str(tmp_path_factory.mktemp('orchestrate_run'))

#     run_orch = lj_orchestrator_defaults.orchestrate_snapshot_run_by_time(start_snaphash,
#                                                                          run_time, n_steps,
#                                                                          work_dir=tempdir)

#     return run_orch

# @pytest.fixture(scope='class')
# def lj_orch_file_orchestrated_run(tmp_path_factory, lj_orchestrator_defaults_file):

#     run_time = 20 # seconds
#     n_steps = 100

#     start_snaphash = lj_orchestrator_defaults_file.get_default_snapshot_hash()

#     tempdir = str(tmp_path_factory.mktemp('orchestrate_run'))

#     run_orch = lj_orchestrator_defaults_file.orchestrate_snapshot_run_by_time(start_snaphash,
#                                                                          run_time, n_steps,
#                                                                          work_dir=tempdir)

#     return run_orch

# @pytest.fixture(scope='class')
# def lj_orch_file_other_orchestrated_run(tmp_path_factory,
#                                         lj_orchestrator_defaults_file_other):

#     run_time = 20 # seconds
#     n_steps = 100

#     start_snaphash = lj_orchestrator_defaults_file_other.get_default_snapshot_hash()

#     tempdir = str(tmp_path_factory.mktemp('orchestrate_run_other'))

#     run_orch = lj_orchestrator_defaults_file_other.orchestrate_snapshot_run_by_time(start_snaphash,
#                                                                          run_time, n_steps,
#                                                                          work_dir=tempdir)

#     return run_orch


# @pytest.fixture(scope='class')
# def lj_orch_reconciled_orchs(tmp_path_factory, lj_apparatus, lj_init_walkers, lj_configuration):

#     run_time = 20 # seconds
#     n_steps = 100

#     # tempdirs for the orchestrators and configuration output
#     first_tempdir = str(tmp_path_factory.mktemp('reconcile_first_run'))
#     second_tempdir = str(tmp_path_factory.mktemp('reconcile_second_run'))

#     first_orch_path = osp.join(first_tempdir, "first.orch.sqlite")
#     second_orch_path = osp.join(second_tempdir, "second.orch.sqlite")

#     # make two orchestrators in their directories
#     first_orch = Orchestrator(orch_path=first_orch_path)
#     second_orch = Orchestrator(orch_path=second_orch_path)

#     # configure them
#     # 1
#     first_orch.set_default_sim_apparatus(lj_apparatus)
#     first_orch.set_default_init_walkers(lj_init_walkers)
#     first_orch.set_default_configuration(lj_configuration)
#     first_orch.gen_default_snapshot()
#     # 2
#     second_orch.set_default_sim_apparatus(lj_apparatus)
#     second_orch.set_default_init_walkers(lj_init_walkers)
#     second_orch.set_default_configuration(lj_configuration)
#     second_orch.gen_default_snapshot()

#     # do independent runs for each of them

#     # start snapshot hashes
#     first_starthash = first_orch.get_default_snapshot_hash()
#     second_starthash = second_orch.get_default_snapshot_hash()

#     # then orchestrate the runs
#     first_run_orch = first_orch.orchestrate_snapshot_run_by_time(first_starthash,
#                                                        run_time, n_steps,
#                                                        work_dir=first_tempdir)

#     second_run_orch = second_orch.orchestrate_snapshot_run_by_time(second_starthash,
#                                                        run_time, n_steps,
#                                                        work_dir=second_tempdir)

#     # then reconcile them
#     reconciled_orch = reconcile_orchestrators(first_run_orch.orch_path, second_run_orch.orch_path)

#     return first_run_orch, second_run_orch, reconciled_orch



"""Routines for converting resampling data to parental lineages.

The core mechanism of weighted ensemble (WE) is to resample cohorts of
parallel simulations. Principally this means 'cloning' and 'merging'
different walkers which gives rise to branching 'family' tree
structures.

The routines in this module are concerned with utilizing raw
resampling records data from a resampler and extracting these lineages
in useful easy to query structures.

Cloning and merging is performed on a cohort of walkers at every
cycle. A walker has both a state and weight. The state is related to
the dynamical state such as the positions, velocities, etc. that
describe the state of a simulation. The weight corresponds to the
probability normalized with the other walkers of the cohort.

An n-clone constitutes copying a state n times to n walkers with w/n
weights where w is the weight of the cloned walker. The cloned walker
is said to be the parent of the new child walkers.

An k-merge constitutes combining k walkers into a single walker with a
weight that is the sum of all k walkers. The state of this walker is
chosen by sampling from the discrete distribution of the k
walkers. The walker that has it's state chosen to persist is referred
to as the 'kept' walker and the other walkers are said to be
'squashed', as they are compressed into a single walker, preserving
their weight but not their state. Squashed walkers are leaves of the
tree and leave no children. The kept walker and walkers for which no
cloning or merging was performed will have a single child.

Routines
--------

resampling_panel : Compiles an unordered collection of resampling records
    into a structured array of records.

parent_panel : Using a parental relationship reduce the records in a
    resampling panel to a structured array of parent indices.

net_parent_table : Reduce the full parent panel (with multiple steps
    per cycle) to the net results of resampling per cycle.

parent_table_discontinuties : Using an interpretation of warping
    records assigns a special value in a parent table for
    discontinuous warping events.

ancestors : Generate the lineage trace of a given walker back through
    history. This is used to get logically contiguous trajectories from
    walker slot trajectories (not necessarily contiguous in storage).

sliding_window : Generate non-redundant sliding window traces over the
    parent forest (tree) imposed over a parent table.

ParentForest : Class that imposes the forest (tree) structure over the
    parent table. Valid for a single contig in the contig tree (forest).
"""


from copy import copy
import itertools as it

import numpy as np
import networkx as nx

DISCONTINUITY_VALUE = -1
"""Special value used to determine if a parent-child relationship has
discontinuous dynamical continuity. Functions in this module uses this
to set this value.

"""

def resampling_panel(resampling_records, is_sorted=False):
    """Converts an unordered collection of resampling records into a
    structured array (lists) corresponding to cycles and resampling
    steps within cycles.

    It is like doing a pivot on the step indices into an extra
    dimension. Hence it can be thought of as a list of tables indexed
    by the cycle, hence the name panel.

    Parameters
    ----------
    resampling_records : list of nametuple records
        A list of resampling records.

    is_sorted : bool
        If this is True it will be assumed that the resampling_records
        are presorted, otherwise they will be sorted.

    Returns
    -------
    resampling_panel : list of list of list of namedtuple records
        The panel (list of tables) of resampling records in order
        (cycle, step, walker)

    """

    res_panel = []

    # if the records are not sorted this must be done:
    if not is_sorted:
        resampling_records.sort()

    # iterate through the resampling records
    rec_it = iter(resampling_records)
    last_cycle_idx = None
    cycle_recs = []
    stop = False
    while not stop:

        # iterate through records until either there is none left or
        # until you get to the next cycle
        cycle_stop = False
        while not cycle_stop:
            try:
                rec = next(rec_it)
            except StopIteration:
                # this is the last record of all the records
                stop = True
                # this is the last record for the last cycle as well
                cycle_stop = True
                # alias for the current cycle
                curr_cycle_recs = cycle_recs
            else:

                # the cycles in the records may not start at 0, but
                # they are in order so we initialize the last
                # cycle_idx so we know when in the records we have
                # gotten to the next cycle of records
                if last_cycle_idx is None:
                    last_cycle_idx = rec.cycle_idx

                # if the resampling record retrieved is from the next
                # cycle we finish the last cycle
                if rec.cycle_idx > last_cycle_idx:
                    cycle_stop = True
                    # save the current cycle as a special
                    # list which we will iterate through
                    # to reduce down to the bare
                    # resampling record
                    curr_cycle_recs = cycle_recs

                    # start a new cycle_recs for the record
                    # we just got
                    cycle_recs = [rec]
                    last_cycle_idx += 1

            if not cycle_stop:
                cycle_recs.append(rec)

            else:

                # we need to break up the records in the cycle into steps
                cycle_table = []

                # temporary container for the step we are working on
                step_recs = []
                step_idx = 0
                step_stop = False
                cycle_it = iter(curr_cycle_recs)
                while not step_stop:
                    try:
                        cycle_rec = next(cycle_it)
                    # stop the step if this is the last record for the cycle
                    except StopIteration:
                        step_stop = True
                        # alias for the current step
                        curr_step_recs = step_recs

                    # or if the next stop index has been obtained
                    else:

                        if cycle_rec.step_idx > step_idx:
                            step_stop = True
                            # save the current step as a special
                            # list which we will iterate through
                            # to reduce down to the bare
                            # resampling record
                            curr_step_recs = step_recs

                            # start a new step_recs for the record
                            # we just got
                            step_recs = [cycle_rec]
                            step_idx += 1


                    if not step_stop:
                        step_recs.append(cycle_rec)
                    else:
                        # go through the walkers for this step since it is completed
                        step_row = [None for _ in range(len(curr_step_recs))]
                        for walker_rec in curr_step_recs:

                            # collect data from the record
                            walker_idx = walker_rec.walker_idx
                            decision_id = walker_rec.decision_id
                            instruction = walker_rec.target_idxs

                            # set the resampling record for the walker in the step records
                            step_row[walker_idx] = (decision_id, instruction)


                        # add the records for this step to the cycle table
                        cycle_table.append(step_row)

        # add the table for this cycles records to the parent panel
        res_panel.append(cycle_table)

    return res_panel


def parent_panel(decision_class, resampling_panel):
    """Using the parental interpretation of resampling records given by
    the decision_class, convert resampling records in a resampling
    panel to parent indices.

    Parameters
    ----------
    decision_class : class implementing Decision interface
        The class that interprets resampling records for parental relationships.
    resampling_panel : list of list of list of namedtuple records
        Panel of resampling records.

    Returns
    -------
    parent_panel : list of list of list of int
        A structured list of the same for as the resampling panel,
        with parent indices swapped for resampling records.

    """

    parent_panel_in = []
    for cycle in resampling_panel:

        # each stage in the resampling for that cycle
        # make a stage parent table
        parent_table = []

        # now iterate through the rest of the stages
        for step in cycle:

            # get the parents idxs for the children of this step
            step_parents = decision_class.parents(step)

            # for the full stage table save all the intermediate parents
            parent_table.append(step_parents)

        # for the full parent panel
        parent_panel_in.append(parent_table)

    return parent_panel_in

def net_parent_table(parent_panel):
    """Reduces a full parent panel to get parent indices on a cycle basis.

    The full parent panel has parent indices for every step in each
    cycle. This computes the net parent relationships for each cycle,
    thus reducing the list of tables (panel) to a single table. A
    table need not have the same length rows, i.e. a ragged array,
    since there can be different numbers of walkers each cycle.

    Parameters
    ----------
    parent_panel : list of list of list of int
        The full panel of parent relationships.

    Returns
    -------
    parent_table : list of list of int
        Net parent relationships for each cycle.

    """

    net_parent_table_in = []

    # each cycle
    for cycle_idx, step_parent_table in enumerate(parent_panel):
        # for the net table we only want the end results,
        # we start at the last cycle and look at its parent
        step_net_parents = []
        n_steps = len(step_parent_table)
        for walker_idx, parent_idx in enumerate(step_parent_table[-1]):
            # initialize the root_parent_idx which will be updated
            root_parent_idx = parent_idx

            # if no resampling skip the loop and just return the idx
            if n_steps > 0:
                # go back through the steps getting the parent at each step
                for prev_step_idx in range(n_steps):
                    prev_step_parents = step_parent_table[-(prev_step_idx+1)]
                    root_parent_idx = prev_step_parents[root_parent_idx]

            # when this is done we should have the index of the root parent,
            # save this as the net parent index
            step_net_parents.append(root_parent_idx)

        # for this step save the net parents
        net_parent_table_in.append(step_net_parents)

    return net_parent_table_in

def parent_table_discontinuities(boundary_condition_class, parent_table, warping_records):
    """Given a parent table and warping records returns a new parent table
    with the discontinuous warping events for parents set to a special
    value (-1).

    Parameters
    ----------
    boundary_condition_class : class implementing BoundaryCondition interface
        The boundary condition class that interprets warping
        records for if they are discontinuous or not.
    parent_table : list of list of int

    warping_records : list of namedtuple records
        The unordered collection of warping events from the simulation.

    Returns
    -------
    parent_table : list of list of int
        Same shape as input parent_table but with discontinuous
        relationships inserted as -1.

    """

    # Make a copy of the parent table
    new_parent_table = copy(parent_table)

    for warp_record in warping_records:

        cycle_idx = warp_record[0]
        parent_idx = warp_record[1]

        n_walkers = len(parent_table[cycle_idx])


        discs = [False for _ in range(n_walkers)]
        # Check to see if any walkers in the current step
        # originated from this warped walker
        for walker_idx in range(n_walkers):

            # this walker was warped discontinuously make a record true for itq
            if boundary_condition_class.warping_discontinuity(warp_record):
                discs[walker_idx] = True

        disc_parent_row = parent_cycle_discontinuities(parent_table[cycle_idx],
                                                       discs)

        new_parent_table[cycle_idx] = disc_parent_row


    return new_parent_table


def parent_cycle_discontinuities(parent_idxs, discontinuities):

    parent_row = copy(parent_idxs)
    for walker_idx, disc in enumerate(discontinuities):

        # if there was a discontinuity in this walker, we need to
        # check for which children it had and apply the discontinuity
        # to them
        if disc:
            for child_idx in range(len(parent_idxs)):

                parent_idx = parent_idxs[child_idx]

                if parent_idx == walker_idx:
                    parent_row[child_idx] = DISCONTINUITY_VALUE

    return parent_row

def ancestors(parent_table, cycle_idx, walker_idx, ancestor_cycle=0):
    """Returns the lineage of ancestors as walker indices leading up to
    the given walker.

    Parameters
    ----------
    parent_table : list of list of int
    cycle_idx : int
        Cycle of walker to query.
    walker_idx : int
        Walker index in to query along with cycle_idx.
    ancestor_cycle : int
        Index of cycle in history to go back to. Must be less than cycle_idx.

    Returns
    -------
    ancestor_trace : list of tuples of ints (traj_idx, cycle_idx)
        Contig walker trace of the ancestors leading up to the queried walker.
        The contig is sequence of cycles in the parent table.

    """

    lineage = [(walker_idx, cycle_idx)]

    previous_walker = walker_idx

    for curr_cycle_idx in range(cycle_idx-1, ancestor_cycle-1, -1):
            previous_walker = parent_table[curr_cycle_idx][previous_walker]

            # check for discontinuities, e.g. warping events
            if previous_walker == -1:
                # there are no more continuous ancestors for this
                # walker so we cannot return ancestors back to the
                # requested cycle just return the ancestors to this
                # point
                break

            previous_point = (previous_walker, curr_cycle_idx)
            lineage.insert(0, previous_point)

    return lineage

def sliding_window(parent_table, window_length):
    """Return contig walker traces of sliding windows of given length over
    the parent forest imposed over the contig given by the parent table.

    Windows are given in no particular order and are nonredundant for
    trees.

    Parameters
    ----------
    parent_table : list of list of int
        Parent table defining parent relationships and contig.
    window_length : int
        Length of window to use. Must be greater than 1.

    Returns
    -------
    windows : list of list of tuples of ints (traj_idx, cycle_idx)
        List of contig walker traces.
    """

    # assert parent_table.dtype == np.int, \
    #     "parent table values must be integers, not {}".format(parent_table.dtype)

    assert window_length > 1, "window length must be greater than one"

    windows = []
    # we make a range iterator which goes from the last cycle to the
    # cycle which would be the end of the first possible sliding window
    for cycle_idx in range(len(parent_table)-1, window_length-2, -1):

        # then iterate for each walker at this cycle
        for walker_idx in range(len(parent_table[0])):

            # then get the ancestors according to the sliding window
            window = ancestors(parent_table, cycle_idx, walker_idx,
                               ancestor_cycle=cycle_idx-(window_length-1))

            # if the window is too short because the lineage has a
            # discontinuity in it skip to the next window
            if len(window) < window_length:
                continue

            windows.append(window)

    return windows


class ParentForest():
    """A tree abstraction to a contig representing the family trees of walkers.

    Uses a directed graph (networkx.DiGraph) to represent parent-child
    relationships; i.e. for edge (A, B) node A is a parent of node B.

    """


    WEIGHT = 'weight'
    """Key for weight node attribute."""
    FREE_ENERGY = 'free_energy'
    """Key for free energy node attribute."""

    ROOT_CYCLE_IDX = -1
    """Special value for the root nodes cycle indices which the initial
     conditions of the simulation

    """
    DISCONTINUITY_VALUE = -1
    """Special value used to determine if a parent-child relationship has
    discontinuous dynamical continuity. This class tests for this
    value in the parent table.
    """

    CONTINUITY_VALUE = 0

    def __init__(self, contig=None,
                 parent_table=None,
    ):
        """Constructs a parent forest from either a Contig object or parent table.

        Either a contig or parent_table must be given but not both.

        The underlying data structure used is a parent table. However,
        if a contig is given a reference to it will be kept.

        Arguments
        ---------
        contig : Conting object, optional conditional on parent_table

        parent_table : list of list of int, optional conditional on contig
            Must not contain the discontinuity values. If you want to
            include metadata on discontinuities use the contig input
            which is preferrable.

        Raises
        ------
        ValueError
            If neither parent_table nor contig is given, or if both are given.

        """


        if (contig is not None) and (parent_table is not None):
            raise ValueError("Pass in either a contig or a parent table but not both")

        if (contig is None) and (parent_table is None):
            raise ValueError("Must provide either a contig or a parent table")

        self._contig = contig

        # if the contig was given, generate a parent table
        if self._contig is not None:

            # from that contig make a parent table
            self._parent_table = self.contig.parent_table(discontinuities=False)

        # otherwise use the one given
        else:
            assert not self.DISCONTINUITY_VALUE in it.chain(*parent_table), \
                "Discontinuity values in parent table are not allowed."

            self._parent_table = parent_table

        self._graph = nx.DiGraph()

        # make the roots of each tree in the parent graph, because
        # this is outside the indexing of the steps we use a special
        # index
        self._roots = [(self.ROOT_CYCLE_IDX, i)
                       for i in range(len(self.parent_table[0]))]

        # set these as nodes
        self.graph.add_nodes_from(self.roots)

        # go through the parent matrix and make edges from the parents
        # to children nodes
        for step_idx, parent_idxs in enumerate(self.parent_table):

            # make edge between each walker of this step to the previous step
            edges, edges_attrs = self._make_child_parent_edges(step_idx, parent_idxs)

            # then put them into this graph
            for i, edge in enumerate(edges):

                if edge is not None:
                    self.graph.add_edge(*edge, **edges_attrs[i])

    def _make_child_parent_edges(self, step_idx, parent_idxs):
        """Generate edge_ids and edge attributes for an array of parent indices.

        Parameters
        ----------
        step_idx : int
            Index of step, just sets the value to put into the node_ids
        parent_idxs : list of int
            For element i in this list, the value j is the index of
            the child in slot i in step_idx+1. Thus j must be
            between 0 and len(parent_idxs)-1.

        Returns
        -------
        edges : list of 2-tuple of node_id
        edge_attributes : list of dict

        """

        edges = []
        edges_attrs = []
        for curr_walker_idx, parent_idx in enumerate(parent_idxs):

            # if the parent is the discontinuity value we set the
            # parent node in the edge as the discontinuity value
            disc = self.CONTINUITY_VALUE
            if parent_idx == self.DISCONTINUITY_VALUE:
                disc = self.DISCONTINUITY_VALUE

            # otherwise we make the edge with the parent and child as
            # normal
            parent_node = (step_idx - 1, parent_idx)
            child_node = (step_idx, curr_walker_idx)

            # make an edge between the parent of this walker and this walker
            edge = (parent_node, child_node)

            edges.append(edge)

            # nothing to do but I already wrote it this way and may be
            # useful later
            edge_attrs = {'discontinuity' : disc}

            edges_attrs.append(edge_attrs)

        return edges, edges_attrs

    @property
    def contig(self):
        """Underlying contig if given during construction."""
        if self._contig is None:
            raise AttributeError("No contig set.")
        else:
            return self._contig

    @property
    def parent_table(self):
        """Underlying parent table data structure."""
        return self._parent_table

    @property
    def graph(self):
        """Underlying networkx.DiGraph object."""
        return self._graph

    @property
    def roots(self):
        """Returns the roots of all the trees in this forest."""
        return self._roots

    @property
    def trees(self):
        """Returns a list of the subtrees from each root in this forest. In no particular order"""
        trees_by_size = [self.graph.subgraph(c) for c in nx.weakly_connected_components(self.graph)]
        trees = []
        for root in self.roots:
            root_tree = [tree for tree in trees_by_size if root in tree][0]
            trees.append(root_tree)
        return trees

    @property
    def n_steps(self):
        """Number of steps of resampling in the parent forest."""

        return len(self.parent_table)

    def step(self, step_idx):
        """Get the nodes at the step (level of the tree).

        Parameters
        ----------
        step_idx : int

        Returns
        -------
        nodes : list of node_id
        """

        step_nodes = []
        for i, node in enumerate(self.graph.nodes):
            if node[0] == step_idx:
                step_nodes.append(node)

        step_nodes.sort()
        return step_nodes

    def steps(self):
        """Returns the nodes ordered by the step and walker indices.

        Returns
        -------
        node_steps : list of list of node_id
        """
        node_steps = []
        for step_idx in range(self.n_steps):
            node_steps.append(self.step(step_idx))

        return node_steps

    def walker(self, walker_idx):
        """Get the nodes for this walker for the whole tree.

        Parameters
        ----------
        walker_idx : int

        Returns
        -------
        nodes : list of node_id

        """

        walker_nodes = []
        for node in self.graph.nodes:
            if node[1] == walker_idx:
                walker_nodes.append(node)

        walker_nodes.sort()
        return walker_nodes

    # TODO DEPRECATE: we shouldn't assume constant number of walkers
    # def walkers(self):
    #     """ """
    #     node_walkers = []
    #     for walker_idx in range(self.n_walkers):
    #         node_walkers.append(self.walker(walker_idx))

    #     return node_walkers

    def set_node_attributes(self, attribute_key, node_attribute_dict):
        """Set attributes for all nodes for a single key.

        Parameters
        ----------
        attribute_key : str
            Key to set for a node attribute.
        node_attribute_dict : dict of node_id: value
            Dictionary mapping nodes to the values that will be set
            for the attribute_key.

        """

        for node_id, value in node_attribute_dict.items():
            self.graph.nodes[node_id][attribute_key] = value

    def set_attrs_by_array(self, attribute_key, values):
        """Set node attributes on a stepwise basis using structural indices.

        Expects a array/list that is n_steps long and has the
        appropriate number of values for the number of walkers at each
        step.

        Parameters
        ----------
        attribute_key : str
            Key to set for a node attribute.
        values : array_like of dim (n_steps, n_walkers) or list of
                 list of values.
            Either an array_like if there is a constant number of
            walkers or a list of lists of the values.
        """
        for step in self.steps():
            for node in step:
                self.graph.nodes[node][attribute_key] = values[node[0]][node[1]]



import os.path as osp
from collections import defaultdict
import itertools as it
import logging
from warnings import warn

from wepy.reporter.dashboard import ResamplerDashboardSection

import numpy as np
import pandas as pd
from tabulate import tabulate

class WExploreDashboardSection(ResamplerDashboardSection):
    RESAMPLER_SECTION_TEMPLATE = \
"""

Resampling Algorithm: {{ name }}

Parameters:
- Max Number of Regions: {{ max_n_regions }}
- Max Region Sizes: {{ max_region_sizes }}

Number of Regions per level:

{{ regions_per_level }}

** Walker Assignments

{{ walker_table }}

** Region Hierarchy

Defined Regions with the number of child regions per parent region:

{{ region_hierarchy }}

** Leaf Region Table

{{ leaf_region_table }}

** WExplore Log

{{ wexplore_log }}

"""

    def __init__(self, resampler=None,
                 max_n_regions=None,
                 max_region_sizes=None,
                 **kwargs
    ):

        if 'name' not in kwargs:
            kwargs['name'] = 'WExploreResampler'

        super().__init__(resampler=resampler,
                         max_n_regions=max_n_regions,
                         max_region_sizes=max_region_sizes,
                         **kwargs
        )

        if resampler is not None:
            self.max_n_regions = resampler.max_n_regions
            self.max_region_sizes = resampler.max_region_sizes
        else:
            assert max_n_regions is not None, \
                "If a resampler is not given must give parameters: max_n_regions"
            assert max_region_sizes is not None, \
                "If a resampler is not given must give parameters: max_n_regions"

            self.max_n_regions = max_n_regions
            self.max_region_sizes = max_region_sizes


        self.n_levels = len(self.max_n_regions)

        # updatables
        self.root_region = ()
        init_leaf_region = tuple([0 for i in range(self.n_levels)])
        self.region_ids = [init_leaf_region]
        self.regions_per_level = []
        self.children_per_region = {}

        # resampling
        self.walker_assignments = []
        self.walker_image_distances = []
        self.curr_region_probabilities = defaultdict(int)
        self.curr_region_counts = defaultdict(int)

        #wexplore
        self.branch_records = []


    def _leaf_regions_to_all_regions(self, region_ids):

        # make a set of all the regions starting with the root region
        regions = set([self.root_region])
        for region_id in region_ids:
            for i in range(len(region_id)):
                regions.add(region_id[0:i+1])

        regions = list(regions)
        regions.sort()

        return regions


    def update_values(self, **kwargs):

        # the region assignments for walkers
        assignments = []
        walker_weights = [walker.weight for walker in kwargs['new_walkers']]
        # re-initialize the current weights dictionary
        self.curr_region_probabilities = defaultdict(int)
        self.curr_region_counts = defaultdict(int)
        for walker_record in kwargs['resampling_data']:

            assignment = tuple(walker_record['region_assignment'])
            walker_idx = walker_record['walker_idx'][0]
            assignments.append((walker_idx, assignment))

            # calculate the probabilities and counts of the regions
            # given the current distribution of walkers
            self.curr_region_probabilities[assignment] += walker_weights[walker_idx]
            self.curr_region_counts[assignment] += 1

        # sort them to get the walker indices in the right order
        assignments.sort()
        # then just get the assignment since it is sorted
        self.walker_assignments = [assignment for walker, assignment in assignments]

        # add to the records for region creation in WExplore
        for resampler_record in kwargs['resampler_data']:

            # get the values
            new_leaf_id = tuple(resampler_record['new_leaf_id'])
            branching_level = resampler_record['branching_level'][0]
            walker_image_distance = resampler_record['distance'][0]

            # add the new leaf id to the list of regions in the order they were created
            self.region_ids.append(new_leaf_id)

            # make a new record for a branching event which is:
            # (region_id, level branching occurred, distance of walker that triggered the branching)
            branch_record = (new_leaf_id,
                             branching_level,
                             walker_image_distance)

            # save it in the records
            self.branch_records.append(branch_record)

        # count the number of child regions each region has
        self.children_per_region = {}
        all_regions = self._leaf_regions_to_all_regions(self.region_ids)
        for region_id in all_regions:
            # if its a leaf region it has no children
            if len(region_id) == self.n_levels:
                self.children_per_region[region_id] = 0

            # all others we cound how many children it has
            else:
                # get all regions that have this one as a root
                children_idxs = set()
                for poss_child_id in all_regions:

                    # get the root at the level of this region for the child
                    poss_child_root = poss_child_id[0:len(region_id)]
                    # if the root is the same we keep it without
                    # counting children below the next level, but we skip the same region
                    if (poss_child_root == region_id) and (poss_child_id != region_id):

                        child_idx = poss_child_id[len(region_id)]

                        children_idxs.add(child_idx)

                # count the children of this region
                self.children_per_region[region_id] = len(children_idxs)

        # count the number of regions at each level
        self.regions_per_level = [0 for i in range(self.n_levels)]
        for region_id, n_children in self.children_per_region.items():
            level = len(region_id)

            # skip the leaves
            if level == self.n_levels:
                continue

            self.regions_per_level[level] += n_children


    def gen_fields(self, **kwargs):

        fields = super().gen_fields(**kwargs)

        regions = self._leaf_regions_to_all_regions(self.region_ids)
        region_children = [self.children_per_region[region] for region in regions]
        region_children_pairs = it.chain(*zip(regions, region_children))
        region_hierarchy = '\n'.join(
            ['{}     {}' for i in range(len(regions))]
        ).format(*region_children_pairs)

        # make a table for the regions
        region_table_colnames = ('region', 'n_walkers', 'curr_weight')
        region_table_d = {}
        region_table_d['region'] = self.region_ids
        region_table_d['n_walkers'] = [self.curr_region_counts[region]
                                       for region in self.region_ids]
        region_table_d['curr_weight'] = [self.curr_region_probabilities[region]
                                         for region in self.region_ids]
        leaf_region_table_df = pd.DataFrame(region_table_d,
                                            columns=region_table_colnames)
        leaf_region_table_df.set_index('region', drop=True)

        leaf_region_table_str = tabulate(leaf_region_table_df,
                                         headers=leaf_region_table_df.columns,
                                         tablefmt='orgtbl')

        # log of branching events
        branching_table_colnames = ('new_leaf_id', 'branching_level', 'trigger_distance')
        branching_table_df = pd.DataFrame(self.branch_records, columns=branching_table_colnames)
        branching_table_str = tabulate(branching_table_df,
                                       headers=branching_table_df.columns,
                                       tablefmt='orgtbl')

        ## walker weights
        walker_weights = [walker.weight for walker in kwargs['new_walkers']]
        # make the table of walkers using pandas, using the order here
        walker_table_colnames = ('weight', 'assignment')
        walker_table_d = {}
        walker_table_d['weight'] = walker_weights
        walker_table_d['assignment'] = self.walker_assignments

        walker_table_df = pd.DataFrame(walker_table_d, columns=walker_table_colnames)
        walker_table_str = tabulate(walker_table_df,
                                    headers=walker_table_df,
                                    tablefmt='orgtbl')


        new_fields = {
            'max_n_regions' : self.max_n_regions,
            'max_region_sizes' : self.max_region_sizes,
            'regions_per_level' : self.regions_per_level,
            'region_hierarchy' : region_hierarchy,
            'leaf_region_table' : leaf_region_table_str,
            'wexplore_log' : branching_table_str,
            'walker_table' : walker_table_str
        }

        fields.update(new_fields)

        return fields




import os.path as osp
import logging
import subprocess
from copy import deepcopy

import click

from wepy.orchestration.orchestrator import (
    reconcile_orchestrators,
    Orchestrator
)

from wepy.reporter.hdf5 import WepyHDF5Reporter
from wepy.hdf5 import WepyHDF5

ORCHESTRATOR_DEFAULT_FILENAME = \
            Orchestrator.ORCH_FILENAME_TEMPLATE.format(config=Orchestrator.DEFAULT_CONFIG_NAME,
                                                       narration=Orchestrator.DEFAULT_NARRATION)

def set_loglevel(loglevel):
    """

    \b
    Parameters
    ----------
    loglevel :
        

    \b
    Returns
    -------

    """

    # try to cast the loglevel as an integer. If that fails interpret
    # it as a string.
    try:
        loglevel_num = int(loglevel)
    except ValueError:
        loglevel_num = getattr(logging, loglevel, None)

    # if no such log level exists in logging the string was invalid
    if loglevel_num is None:
        raise ValueError("invalid log level given")

    logging.basicConfig(level=loglevel_num)

START_HASH = '<start_hash>'
CURDIR = '<curdir>'

def settle_run_options(n_workers=None,
                       job_dir=None,
                       job_name=None,
                       narration=None,
                       monitor_http_port=None,
                       tag=None,
                       configuration=None,
                       start_hash=None,
):
    """

    Parameters
    ----------
    n_workers :
         (Default value = None)
    job_dir :
         (Default value = None)
    job_name :
         (Default value = None)
    narration :
         (Default value = None)

    \b
    Returns
    -------

    """

    # the default for the job name is the start hash if none is given
    if job_name == START_HASH:
        job_name = start_hash

    # if the job_name is given and the default value for the job_dir
    # is given (i.e. not specified by the user) we set the job-dir as
    # the job_name
    if job_name is not None and job_dir == CURDIR:
            job_dir = job_name

    # if the special value for curdir is given we get the systems
    # current directory, this is the default.
    if job_dir == CURDIR:
        job_dir = osp.curdir

    # normalize the job_dir
    job_dir = osp.realpath(job_dir)

    # if a path for a configuration was given we want to use it so we
    # unpickle it and return it, otherwise return None and use the
    # default one in the orchestrator
    config = None
    if configuration is not None:
         with open(configuration, 'rb') as rf:
             config = Orchestrator.deserialize(rf.read())

    ## Monitoring

    # nothing to do, just use the port or tag if its given
    monitor_pkwargs = {
        'tag' : tag,
        'port' : monitor_http_port,
    }

    # we need to reparametrize the configuration here since the
    # orchestrator API will ignore reparametrization values if a
    # concrete Configuration is given.
    if config is not None:

        # if there is a change in the number of workers we need to
        # recalculate all of the partial kwargs

        work_mapper_pkwargs = deepcopy(config.work_mapper_partial_kwargs)

        # if the number of workers has changed update all the relevant
        # fields, otherwise leave it alone
        if work_mapper_pkwargs['num_workers'] != n_workers:

            work_mapper_pkwargs['num_workers'] = n_workers
            work_mapper_pkwargs['device_ids'] = [str(i) for i in range(n_workers)]

        config = config.reparametrize(work_dir=job_dir,
                                      config_name=job_name,
                                      narration=narration,
                                      work_mapper_partial_kwargs=work_mapper_pkwargs,
                                      monitor_partial_kwargs=monitor_pkwargs,
        )

    return job_dir, job_name, narration, config

@click.option('--log', default="WARNING")
@click.option('--n-workers', type=click.INT)
@click.option('--checkpoint-freq', default=None, type=click.INT)
@click.option('--job-dir', default=CURDIR, type=click.Path(writable=True))
@click.option('--job-name', default=START_HASH)
@click.option('--narration', default="")
@click.option('--monitor-http-port', default=9001)
@click.option('--tag', default="None")
@click.argument('n_cycle_steps', type=click.INT)
@click.argument('run_time', type=click.FLOAT)
@click.argument('configuration', type=click.Path(exists=True))
@click.argument('snapshot', type=click.File('rb'))
@click.command()
def run_snapshot(log, n_workers, checkpoint_freq, job_dir, job_name, narration,
                 monitor_http_port, tag,
                 n_cycle_steps, run_time, configuration, snapshot):
    """

    \b
    Parameters
    ----------
    log :
        
    n_workers :
        
    checkpoint_freq :
        
    job_dir :
        
    job_name :
        
    narration :
        
    monitor_http_port :
        
    n_cycle_steps :
        
    run_time :
        
    start_hash :
        
    orchestrator :
        

    \b
    Returns
    -------

    """

    set_loglevel(log)

    logging.info("Loading the starting snapshot file")
    # read the config and snapshot in
    serial_snapshot = snapshot.read()

    logging.info("Creating orchestrating orch database")
    # make the orchestrator for this simulation in memory to start
    orch = Orchestrator()
    logging.info("Adding the starting snapshot to database")
    start_hash = orch.add_serial_snapshot(serial_snapshot)

    # settle what the defaults etc. are for the different options as they are interdependent
    job_dir, job_name, narration, config = settle_run_options(
        # work mapper
        n_workers=n_workers,
        # reporters
        job_dir=job_dir,
        job_name=job_name,
        narration=narration,
        # monitoring
        tag=tag,
        monitor_http_port=monitor_http_port,
        # other
        configuration=configuration,
        start_hash=start_hash,
    )

    # add the parametrized configuration to the orchestrator
    # config_hash = orch.add_serial_configuration(config)

    logging.info("Orchestrator loaded")
    logging.info("Running snapshot by time")
    run_orch = orch.orchestrate_snapshot_run_by_time(start_hash,
                                                     run_time,
                                                     n_cycle_steps,
                                                     checkpoint_freq=checkpoint_freq,
                                                     work_dir=job_dir,
                                                     config_name=job_name,
                                                     narration=narration,
                                                     configuration=config,
                                                     )
    logging.info("Finished running snapshot by time")

    start_hash, end_hash = run_orch.run_hashes()[0]

    run_orch.close()
    logging.info("Closed the resultant orch")

    # write the run tuple out to the log
    run_line_str = "Run start and end hashes: {}, {}".format(start_hash, end_hash)

    # log it
    logging.info(run_line_str)

    # also put it to the terminal
    click.echo(run_line_str)

    orch.close()
    logging.info("closed the orchestrating orch database")

@click.option('--log', default="WARNING")
@click.option('--n-workers', type=click.INT)
@click.option('--checkpoint-freq', default=None, type=click.INT)
@click.option('--job-dir', default=CURDIR, type=click.Path(writable=True))
@click.option('--job-name', default=START_HASH)
@click.option('--narration', default="")
@click.option('--configuration', type=click.Path(exists=True), default=None)
@click.argument('n_cycle_steps', type=click.INT)
@click.argument('run_time', type=click.FLOAT)
@click.argument('start_hash')
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def run_orch(log, n_workers, checkpoint_freq, job_dir, job_name, narration, configuration,
        n_cycle_steps, run_time, start_hash, orchestrator):
    """

    \b
    Parameters
    ----------
    log :
        
    n_workers :
        
    checkpoint_freq :
        
    job_dir :
        
    job_name :
        
    narration :
        
    n_cycle_steps :
        
    run_time :
        
    start_hash :
        
    orchestrator :
        

    \b
    Returns
    -------

    """

    set_loglevel(log)

    # settle what the defaults etc. are for the different options as they are interdependent
    job_dir, job_name, narration, config = settle_run_options(n_workers=n_workers,
                                                                         job_dir=job_dir,
                                                                         job_name=job_name,
                                                                         narration=narration,
                                                                         configuration=configuration,
                                                                         start_hash=start_hash)

    # Open a wrapper around the orchestrator database that provides
    # the inputs for the simulation
    orch = Orchestrator(orchestrator, mode='r')

    logging.info("Orchestrator loaded")

    logging.info("Running snapshot by time")
    run_orch = orch.orchestrate_snapshot_run_by_time(start_hash,
                                                     run_time,
                                                     n_cycle_steps,
                                                     checkpoint_freq=checkpoint_freq,
                                                     work_dir=job_dir,
                                                     config_name=job_name,
                                                     narration=narration,
                                                     configuration=config,
                                                     )
    logging.info("Finished running snapshot by time")

    start_hash, end_hash = run_orch.run_hashes()[0]

    logging.info("Closing the resultant orchestrator")
    run_orch.close()

    # write the run tuple out to the log
    run_line_str = "Run start and end hashes: {}, {}".format(start_hash, end_hash)

    # log it
    logging.info(run_line_str)

    # also put it to the terminal
    click.echo(run_line_str)

    logging.info("Closing the orchestrating orch")
    orch.close()

def combine_orch_wepy_hdf5s(new_orch, new_hdf5_path, run_ids=None):
    """

    \b
    Parameters
    ----------
    new_orch :
        
    new_hdf5_path :
        

    \b
    Returns
    -------

    """

    if run_ids is None:
        run_ids = new_orch.run_hashes()

    # we assume that the run we are interested in is the only run in
    # the WepyHDF5 file so it is index 0
    singleton_run_idx = 0

    # a key-value for the paths for each run
    hdf5_paths = {}

    # go through each run in the new orchestrator
    for run_id in run_ids:

        # get the configuration used for this run
        run_config = new_orch.run_configuration(*run_id)

        # from that configuration find the WepyHDF5Reporters
        for reporter in run_config.reporters:

            if isinstance(reporter, WepyHDF5Reporter):

                # and save the path for that run
                hdf5_paths[run_id] = reporter.file_path

    click.echo("Combining these HDF5 files:")
    click.echo('\n'.join(hdf5_paths.values()))

    # now that we have the paths (or lack of paths) for all
    # the runs we need to start linking them all
    # together.

    # first we need a master linker HDF5 to do this with

    # so load a template WepyHDF5
    template_wepy_h5_path = hdf5_paths[run_ids[singleton_run_idx]]
    template_wepy_h5 = WepyHDF5(template_wepy_h5_path, mode='r')

    # clone it
    with template_wepy_h5:
        master_wepy_h5 = template_wepy_h5.clone(new_hdf5_path, mode='x')

    click.echo("Into a single master hdf5 file: {}".format(new_hdf5_path))

    # then link all the files to it
    run_mapping = {}
    for run_id, wepy_h5_path in hdf5_paths.items():

        # in the case where continuations were done from
        # checkpoints then the runs data will potentially (and
        # most likely) contain extra cycles since checkpoints are
        # typically produced on some interval of cycles. So, in
        # order for us to actually piece together contigs we need
        # to take care of this.

        # There are two ways to deal with this which can both be
        # done at the same time. The first is to keep the "nubs",
        # which are the small leftover pieces after the checkpoint
        # that ended up getting continued, and make a new run from
        # the last checkpoint to the end of the nub, in both the
        # WepyHDF5 and the orchestrator run collections.

        # The second is to generate a WepyHDF5 run that
        # corresponds to the run in the checkpoint orchestrator.

        # To avoid complexity (for now) we opt to simply dispose
        # of the nubs and assume that not much will be lost from
        # this. For the typical use case of making multiple
        # independent and linear contigs this is also the simplest
        # mode, since the addition of multiple nubs will introduce
        # an extra spanning contig in the contig tree.

        # furthermore the nubs provide a source of problems if
        # rnus were abruptly stopped and data is not written some
        # of the frames can be corrupted. SO until we know how to
        # stop this (probably SWMR mode will help) this is also a
        # reason not to deal with nubs.

        # TODO: add option to keep nubs in HDF5, and deal with in
        # orch (you won't be able to have an end snapshot...).

        # to do this we simply check whether or not the number of
        # cycles for the run_id are less than the number of cycles
        # in the corresponding WepyHDF5 run dataset.
        orch_run_num_cycles = new_orch.run_last_cycle_idx(*run_id)

        # get the number of cycles that are in the data for the run in
        # the HDF5 to compare to the number in the orchestrator run
        # record
        wepy_h5 = WepyHDF5(wepy_h5_path, mode='r')
        with wepy_h5:
            h5_run_num_cycles = wepy_h5.num_run_cycles(singleton_run_idx)

        # sanity check for if the number of cycles in the
        # orchestrator is greater than the HDF5
        if orch_run_num_cycles > h5_run_num_cycles:
            raise ValueError("Number of cycles in orch run is more than HDF5."\
                             "This implies missing data")

        # copy the run (with the slice)
        with master_wepy_h5:

            # TODO: this was the old way of combining where we would
            # just link, however due to the above discussion this is
            # not tenable now. In the future there might be some more
            # complex options taking linking into account but for now
            # we just don't use it and all runs will be copied by this
            # operation

            # # we just link the whole file then sort out the
            # # continuations later since we aren't necessarily doing
            # # this in a logical order
            # new_run_idxs = master_wepy_h5.link_file_runs(wepy_h5_path)

            # extract the runs from the file (there should only be
            # one). This means copy the run, but if we only want a
            # truncation of it we will use the run slice to only get
            # part of it

            # so first we generate the run slices for this file using
            # the number of cycles recorded in the orchestrator
            run_slices = {singleton_run_idx : (0, orch_run_num_cycles)}

            click.echo("Extracting Run: {}".format(run_id))
            click.echo("Frames 0 to {} out of {}".format(
                orch_run_num_cycles, h5_run_num_cycles))

            # then perform the extraction, which will open the other
            # file on its own
            new_run_idxs = master_wepy_h5.extract_file_runs(wepy_h5_path,
                                                            run_slices=run_slices)

            # map the hash id to the new run idx created. There should
            # only be one run in an HDF5 if we are following the
            # orchestration workflow.
            assert len(new_run_idxs) < 2, \
                "Cannot be more than 1 run per HDF5 file in orchestration workflow"

            run_mapping[run_id] = new_run_idxs[0]

            click.echo("Set as run: {}".format(new_run_idxs[0]))

    click.echo("Done extracting runs, setting continuations")

    with master_wepy_h5:

        # now that they are all linked we need to add the snapshot
        # hashes identifying the runs as metadata. This is so we can
        # map the simple run indices in the HDF5 back to the
        # orchestrator defined runs. This will be saved as metadata on
        # the run. Also:

        # We need to set the continuations correctly betwen the runs
        # in different files, so for each run we find the run it
        # continues in the orchestrator
        for run_id, run_idx in run_mapping.items():

            # set the run snapshot hash metadata except for if we have
            # already done it
            try:
                master_wepy_h5.set_run_start_snapshot_hash(run_idx, run_id[0])
            except AttributeError:
                # it was already set so just move on
                pass
            try:
                master_wepy_h5.set_run_end_snapshot_hash(run_idx, run_id[1])
            except AttributeError:
                # it was already set so just move on
                pass

            # find the run_id that this one continues
            continued_run_id = new_orch.run_continues(*run_id)

            # if a None is returned then there was no continuation
            if continued_run_id is None:
                # so we go to the next run_id and don't log any
                # continuation
                continue

            # get the run_idx in the HDF5 that corresponds to this run
            continued_run_idx = run_mapping[continued_run_id]

            click.echo("Run {} continued by {}".format(continued_run_id, run_idx))

            # add the continuation
            master_wepy_h5.add_continuation(run_idx, continued_run_idx)

@click.command()
@click.argument('orchestrator', nargs=1, type=click.Path(exists=True))
@click.argument('hdf5', nargs=1, type=click.Path(exists=False))
@click.argument('run_ids', nargs=-1)
def reconcile_hdf5(orchestrator, hdf5, run_ids):
    """For an orchestrator with multiple runs combine the HDF5 results
    into a single one.

    This requires that the paths inside of the reporters for the
    configurations used for a run still have valid paths to the HDF5
    files.

    \b
    Parameters
    ----------
    orchestrator : Path
        The orchestrator to retrieve HDF5s for

    hdf5 : Path
        Path to the resultant HDF5.

    run_ids : str
        String specifying a run as start and end hash
        e.g. 'd0cb2e6fbcc8c2d66d67c845120c7f6b,b4b96580ae57f133d5f3b6ce25affa6d'

    \b
    Returns
    -------

    """

    # parse the run ids
    run_ids = [tuple(run_id.split(',')) for run_id in run_ids]

    orch = Orchestrator(orchestrator, mode='r')

    hdf5_path = osp.realpath(hdf5)

    click.echo("Combining the HDF5s together, saving to:")
    click.echo(hdf5_path)

    # combine the HDF5 files from those orchestrators
    combine_orch_wepy_hdf5s(orch, hdf5_path, run_ids=run_ids)


@click.command()
@click.option('--hdf5', type=click.Path(exists=False))
@click.argument('output', nargs=1, type=click.Path(exists=False))
@click.argument('orchestrators', nargs=-1, type=click.Path(exists=True))
def reconcile_orch(hdf5, output, orchestrators):
    """ 

    \b
    Parameters
    ----------
    hdf5 : Path
        Path to the resultant HDF5.
    output : Path
        Path to the resultant orchestrator that is created.
    orchestrators : Path
        Paths to the orchestrators to reconcile.

    \b
    Returns
    -------

    """

    new_orch = reconcile_orchestrators(output, *orchestrators)

    # if a path for an HDF5 file is given
    if hdf5 is not None:
        hdf5_path = osp.realpath(hdf5)

        click.echo("Combining the HDF5s together, saving to:")
        click.echo(hdf5_path)

        # combine the HDF5 files from those orchestrators
        combine_orch_wepy_hdf5s(new_orch, hdf5_path)



def hash_listing_formatter(hashes):
    """

    \b
    Parameters
    ----------
    hashes :
        

    \b
    Returns
    -------

    """
    hash_listing_str = '\n'.join(hashes)
    return hash_listing_str

@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def ls_snapshots(orchestrator):
    """

    \b
    Parameters
    ----------
    orchestrator :
        

    \b
    Returns
    -------

    """

    orch = Orchestrator(orch_path=orchestrator, mode='r')

    message = hash_listing_formatter(orch.snapshot_hashes)

    orch.close()

    click.echo(message)

@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def ls_runs(orchestrator):
    """

    \b
    Parameters
    ----------
    orchestrator :
        

    \b
    Returns
    -------

    """

    orch = Orchestrator(orch_path=orchestrator, mode='r')

    runs = orch.run_hashes()

    orch.close()

    hash_listing_str = "\n".join(["{}, {}".format(start, end) for start, end in runs])

    click.echo(hash_listing_str)

@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def ls_configs(orchestrator):
    """

    \b
    Parameters
    ----------
    orchestrator :
        

    \b
    Returns
    -------

    """

    orch = Orchestrator(orch_path=orchestrator, mode='r')

    message = hash_listing_formatter(orch.configuration_hashes)

    orch.close()

    click.echo(message)



@click.command()
@click.option('--no-expand-external', is_flag=True)
@click.argument('source', type=click.Path(exists=True))
@click.argument('target', type=click.Path(exists=False))
def hdf5_copy(no_expand_external, source, target):
    """Copy a WepyHDF5 file, except links to other runs will optionally be
    expanded and truly duplicated if symbolic inter-file links are present."""


    # arg clusters to pass to subprocess for the files
    input_f_args = ['-i', source]
    output_f_args = ['-o', target]

    # each invocation calls a different group since we can't call the
    # toplevel '/' directly
    settings_args = ['-s', '/units', '-d', '/units']
    settings_args = ['-s', '/_settings', '-d', '/_settings']
    topology_args = ['-s', '/topology', '-d', '/topology']
    runs_args = ['-s', '/runs', '-d', '/runs']

    # by default expand the external links
    flags_args = ['-f', 'ext']

    # if the not expand external flag is given get rid of those args
    if no_expand_external:
        flags_args = []

    common_args = input_f_args + output_f_args + flags_args

    settings_output = subprocess.check_output(['h5copy'] + common_args + settings_args)

    topology_output = subprocess.check_output(['h5copy'] + common_args + topology_args)

    runs_output = subprocess.check_output(['h5copy'] + common_args + runs_args)


@click.option('-O', '--output', type=click.Path(exists=False), default=None)
@click.argument('snapshot_hash')
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def get_snapshot(output, snapshot_hash, orchestrator):

    # first check if the output is None, if it is we automatically
    # generate a file in the cwd that is the hash of the snapshot
    if output is None:
        output = "{}.snap.dill.pkl".format(snapshot_hash)

        # check that it doesn't exist, and fail if it does, since we
        # don't want to implicitly overwrite stuff
        if osp.exists(output):
            raise OSError("No output path was specified and default alredy exists, exiting.")

    orch = Orchestrator(orchestrator, mode='r')

    serial_snapshot = orch.snapshot_kv[snapshot_hash]

    with open(output, 'wb') as wf:
        wf.write(serial_snapshot)

    orch.close()

@click.option('-O', '--output', type=click.Path(exists=False), default=None)
@click.argument('config_hash')
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def get_config(output, config_hash, orchestrator):

    # first check if the output is None, if it is we automatically
    # generate a file in the cwd that is the hash of the snapshot
    if output is None:
        output = "{}.config.dill.pkl".format(config_hash)

        # check that it doesn't exist, and fail if it does, since we
        # don't want to implicitly overwrite stuff
        if osp.exists(output):
            raise OSError("No output path was specified and default alredy exists, exiting.")

    orch = Orchestrator(orchestrator, mode='r')

    serial_snapshot = orch.configuration_kv[config_hash]

    with open(output, 'wb') as wf:
        wf.write(serial_snapshot)

        orch.close()

@click.option('-O', '--output', type=click.Path(exists=False), default=None)
@click.argument('end_hash')
@click.argument('start_hash')
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def get_run(output, end_hash, start_hash, orchestrator):

    # first check if the output is None, if it is we automatically
    # generate a file in the cwd that is the hash of the snapshot
    if output is None:
        output = "{}-{}.orch.sqlite".format(start_hash, end_hash)

        # check that it doesn't exist, and fail if it does, since we
        # don't want to implicitly overwrite stuff
        if osp.exists(output):
            raise OSError("No output path was specified and default alredy exists, exiting.")

    orch = Orchestrator(orchestrator, mode='r')

    start_serial_snapshot = orch.snapshot_kv[start_hash]
    end_serial_snapshot = orch.snapshot_kv[end_hash]


    # get the records values for this run
    rec_d = {field : value for field, value in
           zip(Orchestrator.RUN_SELECT_FIELDS, orch.get_run_record(start_hash, end_hash))}

    config = orch.configuration_kv[rec_d['config_hash']]

    # create a new orchestrator at the output location
    new_orch = Orchestrator(output, mode='w')

    _ = new_orch.add_serial_snapshot(start_serial_snapshot)
    _ = new_orch.add_serial_snapshot(end_serial_snapshot)
    config_hash = new_orch.add_serial_configuration(config)

    new_orch.register_run(start_hash, end_hash, config_hash, rec_d['last_cycle_idx'])

    orch.close()
    new_orch.close()

@click.argument('end_hash')
@click.argument('start_hash')
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def get_run_cycles(end_hash, start_hash, orchestrator):

    orch = Orchestrator(orchestrator, mode='r')

    start_serial_snapshot = orch.snapshot_kv[start_hash]
    end_serial_snapshot = orch.snapshot_kv[end_hash]

    # get the records values for this run
    rec_d = {field : value for field, value in
           zip(Orchestrator.RUN_SELECT_FIELDS, orch.get_run_record(start_hash, end_hash))}

    click.echo(rec_d['last_cycle_idx'])


@click.argument('orchestrator', type=click.Path(exists=False))
@click.command()
def create_orch(orchestrator):

    orch = Orchestrator(orchestrator, mode='x')

    orch.close()

@click.argument('snapshot', type=click.File('rb'))
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def add_snapshot(snapshot, orchestrator):

    orch = Orchestrator(orchestrator, mode='r+')

    serial_snapshot = snapshot.read()

    snaphash = orch.add_serial_snapshot(serial_snapshot)

    orch.close()

    click.echo(snaphash)


@click.argument('configuration', type=click.File('rb'))
@click.argument('orchestrator', type=click.Path(exists=True))
@click.command()
def add_config(configuration, orchestrator):
    orch = Orchestrator(orchestrator, mode='r+')

    serial_config = configuration.read()

    config_hash = orch.add_serial_snapshot(serial_config)

    orch.close()

    click.echo(config_hash)


@click.group()
def cli():
    """ """
    pass

@click.group()
def run():
    """ """
    pass

@click.group()
def get():
    """ """
    pass

@click.group()
def add():
    """ """
    pass

@click.group()
def create():
    """ """
    pass


@click.group()
def ls():
    """ """
    pass

@click.group()
def reconcile():
    """ """
    pass

@click.group()
def hdf5():
    """ """
    pass

# command groupings

# run
run.add_command(run_orch, name='orch')
run.add_command(run_snapshot, name='snapshot')

# ls
ls.add_command(ls_snapshots, name='snapshots')
ls.add_command(ls_runs, name='runs')
ls.add_command(ls_configs, name='configs')

# get
get.add_command(get_snapshot, name='snapshot')
get.add_command(get_config, name='config')
get.add_command(get_run, name='run')
get.add_command(get_run_cycles, name='run-cycles')

# add
add.add_command(add_snapshot, name='snapshot')
add.add_command(add_config, name='config')

# create
create.add_command(create_orch, name='orch')

# reconcile
reconcile.add_command(reconcile_orch, name='orch')
reconcile.add_command(reconcile_hdf5, name='hdf5')

# hdf5
hdf5.add_command(hdf5_copy, name='copy')
# desired commands
# hdf5.add_command(hdf5_copy, name='copy-run')
# hdf5.add_command(hdf5_copy, name='copy-traj')
# hdf5.add_command(hdf5_copy, name='ls-runs')
# hdf5.add_command(hdf5_copy, name='ls-run-hashes')

# subgroups
subgroups = [run, get, add, create, ls, reconcile, hdf5]

for subgroup in subgroups:
    cli.add_command(subgroup)

if __name__ == "__main__":

    cli()



"""Analysis tools for wepy datasets.

Analysis of weighted ensemble (WE) datasets are complicated by their
nonlinear branching structure, and thus in its raw form cannot be
analyzed by the tools available for linear trajectories. See the
documentation (TODO: link) for a more in depth explanation of the
general structure of weighted ensemble data.

These analysis modules do not aim to redo any particular analysis
calculation but rather to aid in accessing, summarizing, and
manipulating branched simulation data.

The starting point for this is the `ContigTree` class in the
`contig_tree` module. This is a general purpose container that
combines the data specifying the branching structure with the
underlying linear data structures. It is necessary for branched data
but can also be used for collections of linear data as well.

The main anlaysis routines we attempt to facilitate are:

1. Generating probability histograms over calculated
   observables. (TODO)

2. Generating contiguous trajectories of walkers that have reached
   target boundary conditions (TODO).

3. Visualizing the cloning and merging history of simulations
   (i.e. resampling trees) using the `parents` module and
   `network_layouts`.

4. Generating 'state networks' from frame labels and computing
   transition matrices with variable lag times using the `network` and
   `transitions` modules.

To calculate observables on the dataset you need to work directly with
the data storage API in the `wepy.hdf5.WepyHDF5` module using the
`compute_observable` method.

See Also
--------

`wepy.hdf5.WepyHDF5.compute_observable`

Notes
-----

References
----------

Examples
--------

"""



"""Module for the main simulation management class.

All component class interfaces are set by how the manager interacts
with them.

Managers should implement a three phase protocol for running
simulations:

- init
- run_simulation
- cleanup

The separate `init` method is different than the constructor
`__init__` method and instead calls the special `init` method on all
wepy components (runner, resampler, boundary conditions, and
reporters) at runtime.

This allows for a things that need to be done at runtime before a
simulation begins, e.g. opening files, that you don't want done at
construction time.

This is useful for orchestration because the complete simulation
'image' can be made before runtime (and pickled or otherwise
persisted) without producing external effects.

This is used primarily for reporters, which perform I/O, and work
mappers which may spawn processes.

The `cleanup` method should be called either when the simulation ends
normally as well as when the simulation ends abnormally.

This allows file handles to be closed and processes to be killed at
the end of a simulation and upon failure.


The base methods for running simulations is `run_cycle` which runs
a cycle of weighted ensemble given the state of all the components.

The simulation manager should provide multiple ways of running
simulations depending on if the number of cycles is known up front or
to be determined adaptively (e.g. according to some time limit).

"""

import numpy as np

import sys
import time
from copy import deepcopy
import logging

from eliot import start_action, log_call

from wepy.work_mapper.mapper import Mapper

class Manager(object):
    """The class that coordinates wepy simulations.

    The Manager class is the lynchpin of wepy simulations and is where
    all the different components are composed.

    Strictly speaking the Manager defines the interfaces each
    component must provide to function.

    Developers can call `run_cycle` directly but the following
    convenience functions are provided to run many cycles in
    succession as a single 'run' with consecutive cycle idxs:

    - run_simulation_by_time
    - run_simulation

    The corresponding 'continue' run methods will simply pass a run
    index to reporters indicating that the run continues another.

    For these run methods the `init` method is called followed by
    iterative calls to `run_cycle` and finally with a call to
    `cleanup`.

    The order of application of wepy components are:

    - runner
    - boundary_conditions
    - resampler
    - reporters

    """


    REPORT_ITEM_KEYS = ('cycle_idx',
                        'n_segment_steps',
                        'new_walkers',
                        'resampled_walkers',
                        'warp_data',
                        'bc_data',
                        'progress_data',
                        'resampling_data',
                        'resampler_data',
                        'worker_segment_times',
                        'cycle_runner_time',
                        'cycle_bc_time',
                        'cycle_resampling_time',
    )
    """Keys of values that will be passed to reporters.

    This indicates the values that the reporters will have access to.
    """

    def __init__(self, init_walkers,
                 runner = None,
                 work_mapper = None,
                 resampler = None,
                 boundary_conditions = None,
                 reporters = None,
                 sim_monitor = None,
    ):
        """Constructor for Manager.

        Arguments
        ---------

        init_walkers : list of walkers
            The list of the initial walkers that will be run.

        runner : object implementing the Runner interface
            The runner to be used for propagating sampling segments of walkers.

        work_mapper : object implementing the WorkMapper interface
            The object that will be used to perform a set of runner
            segments in a cycle.

        resampler : object implementing the Resampler interface
            The resampler to be used in the simulation

        boundary_conditions : object implementing BoundaryCondition interface, optional
            The boundary conditions to apply to walkers

        reporters : list of objects implenting the Reporter interface, optional
            Reporters to be used. You should provide these if you want to keep data.

        sim_monitor : SimMonitor object
            An object implementing SimMonitor interface for providing
            monitoring metrics of a simulation.

        Warnings
        --------

        While reporters are strictly optional, you probably want to
        provide some because the simulation manager provides no
        utilities for saving data from the simulations except for the
        walkers at the end of a cycle or simulation.

        See Also
        --------
        wepy.reporter.hdf5 : The standard reporter for molecular simulations in wepy.

        wepy.orchestration.orchestrator.Orchestrator : for running simulations with
            checkpointing, restarting, reporter localization, and configuration hotswapping
            with command line interface.

        """

        self.init_walkers = init_walkers
        self.n_init_walkers = len(init_walkers)

        # the runner is the object that runs dynamics
        self.runner = runner
        # the resampler
        self.resampler = resampler
        # object for boundary conditions
        self.boundary_conditions = boundary_conditions

        # the method for writing output
        if reporters is None:
            self.reporters = []
        else:
            self.reporters = reporters

        if work_mapper is None:
            self.work_mapper = Mapper()
        else:
            self.work_mapper = work_mapper

        ## Monitor
        self.monitor = sim_monitor

        # used to have a record of the last report for the simulation
        # monitor without breaking the API. Ugly but I don't want to
        # break it and no one cares about this anyhow
        self._last_report = None

    @log_call(
        include_args=[
            'segment_length',
            'cycle_idx',
        ],
        include_result=False,
    )
    def run_segment(self, walkers, segment_length, cycle_idx):
        """Run a time segment for all walkers using the available workers.

        Maps the work for running each segment for each walker using
        the work mapper.

        Walkers will have the same weights but different states.

        Parameters
        ----------
        walkers : list of walkers
        segment_length : int
            Number of steps to run in each segment.

        Returns
        -------

        new_walkers : list of walkers
           The walkers after the segment of sampling simulation.

        """

        num_walkers = len(walkers)

        logging.info("Starting segment")

        try:
            new_walkers = list(self.work_mapper.map(
                # args, which must be supported by the map function
                walkers,
                (segment_length for i in range(num_walkers)),

                # kwargs which are optionally recognized by the map function
                cycle_idx=(cycle_idx for i in range(num_walkers)),
                walker_idx=(walker_idx for walker_idx in range(num_walkers)),
            )
            )

        except Exception as exception:

            # get the errors from the work mapper error queue
            self.cleanup()

            # report on all of the errors that occured
            raise exception

        logging.info("Ending segment")

        return new_walkers

    @log_call(include_args=[
        'n_segment_steps',
        'cycle_idx',
        'runner_opts'],
              include_result=False)
    def run_cycle(self,
                  walkers,
                  n_segment_steps,
                  cycle_idx,
                  runner_opts=None,
    ):
        """Run a full cycle of weighted ensemble simulation using each
        component.

        The order of application of wepy components are:

        - runner
        - boundary_conditions
        - resampler
        - reporters

        The `init` method should have been called before this or
        components may fail.

        This method is not idempotent and will alter the state of wepy
        components.

        The cycle is not kept as a state variable of the simulation
        manager and so myst be provided here. This motivation for this
        is that a cycle index is really a property of a run and runs
        can be composed in many ways and is then handled by
        higher-level methods calling run_cycle.

        Each component should implement its respective interface to be
        called in this method, the order and names of the methods
        called are as follows:

        1. runner.pre_cycle
        2. run_segment -> work_mapper.map(runner.run_segment)
        3. runner.post_cycle
        4. boundary_conditions.warp_walkers (if present)
        5. resampler.resample
        6. reporter.report for all reporters

        The pre and post cycle calls to the runner allow for a parity
        of one call per cycle to the runner.

        The boundary_conditions component is optional, as are the
        reporters (although it won't be very useful to run this
        without any).

        Parameters
        ----------
        walkers : list of walkers

        n_segment_steps : int
            Number of steps to run in each segment.

        cycle_idx : int
            The index of this cycle.

        Returns
        -------

        new_walkers : list of walkers
            The resulting walkers of the cycle

        sim_components : list
            The runner, resampler, and boundary conditions
            objects at the end of the cycle.

        See Also
        --------
        run_simulation : To run a simulation by the number of cycles
        run_simulation_by_time

        """


        # this one is called to just easily be able to catch all the
        # errors from it so we can cleanup if an error is caught


        return self._run_cycle(
            walkers,
            n_segment_steps,
            cycle_idx,
            runner_opts=runner_opts,
        )

    def _run_cycle(self,
                   walkers,
                   n_segment_steps,
                   cycle_idx,
                   runner_opts=None,
    ):
        """See run_cycle."""

        if runner_opts is None:
            runner_opts = {}

        # run the runner pre-cycle hook

        start = time.time()

        self.runner.pre_cycle(
            walkers=walkers,
            n_segment_steps=n_segment_steps,
            cycle_idx=cycle_idx,
            **runner_opts
        )

        end = time.time()
        runner_precycle_time = end - start

        # run the segment
        start = time.time()

        new_walkers = self.run_segment(walkers, n_segment_steps, cycle_idx)

        end = time.time()
        sim_manager_segment_time = end - start

        if hasattr(self.runner, '_last_cycle_segments_split_times'):

            runner_splits = deepcopy(self.runner._last_cycle_segments_split_times)

        else:
            runner_splits = None

        logging.info("Starting post cycle")
        # run post-cycle hook
        start = time.time()

        self.runner.post_cycle()

        end = time.time()
        runner_postcycle_time = end - start

        logging.info("End cycle {}".format(cycle_idx))

        # boundary conditions should be optional;

        # initialize the warped walkers to the new_walkers and
        # change them later if need be
        warped_walkers = new_walkers
        warp_data = []
        bc_data = []
        progress_data = {}
        bc_time = 0.0
        if self.boundary_conditions is not None:

            # apply rules of boundary conditions and warp walkers through space
            start = time.time()
            logging.info("Starting boundary conditions")
            bc_results  = self.boundary_conditions.warp_walkers(new_walkers,
                                                                cycle_idx)
            end = time.time()
            bc_time = end - start

            # warping results
            warped_walkers = bc_results[0]
            warp_data = bc_results[1]
            bc_data = bc_results[2]
            progress_data = bc_results[3]

            if len(warp_data) > 0:
                logging.info("Returned warp record in cycle {}".format(cycle_idx))


        # resample walkers
        start = time.time()
        logging.info("Starting resampler")

        resampling_results = self.resampler.resample(warped_walkers)

        end = time.time()
        resampling_time = end - start

        resampled_walkers = resampling_results[0]
        resampling_data = resampling_results[1]
        resampler_data = resampling_results[2]

        # log the weights of the walkers after resampling

        # DEBUG: commenting this out to make mocking easier. But really I
        # don't even care if it stays. as its mostly noise.
        # result_template_str = "|".join(["{:^5}" for i in range(self.n_init_walkers + 1)])
        # walker_weight_str = result_template_str.format("weight",
        #     *[round(walker.weight, 3) for walker in resampled_walkers])
        # logging.info(walker_weight_str)

        # make a dictionary of all the results that will be reported
        seg_times = {}
        sampling_time = None
        overhead_time = None

        if hasattr(self.work_mapper, 'worker_segment_times'):

            seg_times = deepcopy(self.work_mapper.worker_segment_times)

            # count up the total sampling time from the segments
            sampling_time = 0
            for worker_id, segments_times in self.work_mapper.worker_segment_times.items():
                for seg_time in segments_times:
                    sampling_time += seg_time

            # calculate the overhead for logging
            sim_manager_segment_overhead_time = sim_manager_segment_time - sampling_time

            # logging.info(
            #     "Runner time = {}; Sampling = ({}); Overhead = ({})".format(
            #         runner_time, sampling_time, overhead_time))

        else:
            sim_manager_segment_overhead_time = 0.
            # logging.info("No worker segment times given")
            # logging.info("Runner time = {}".format(runner_time))


        report = {'cycle_idx' : cycle_idx,
                  'new_walkers' : new_walkers,
                  'warp_data' : warp_data,
                  'bc_data' : bc_data,
                  'progress_data' : progress_data,
                  'resampling_data' : resampling_data,
                  'resampler_data' : resampler_data,
                  'n_segment_steps' : n_segment_steps,
                  'resampled_walkers' : resampled_walkers,

                  # timings
                  'runner_precycle_time' : runner_precycle_time,
                  'runner_postcycle_time' : runner_postcycle_time,
                  'sim_manager_segment_overhead_time' : sim_manager_segment_overhead_time,
                  'runner_splits_time' : runner_splits,
                  'worker_segment_times' : seg_times,
                  'cycle_sim_manager_segment_time' : sim_manager_segment_time,
                  'cycle_runner_time' : sim_manager_segment_time,
                  'cycle_bc_time' : bc_time,
                  'cycle_resampling_time' : resampling_time,

        }

        self._last_report = report

        # check that all of the keys that are specified for this sim
        # manager are present
        assert all([True if rep_key in report else False
                    for rep_key in self.REPORT_ITEM_KEYS])

        logging.info("Starting reporting")
        # report results to the reporters
        for reporter in self.reporters:
            reporter.report(**report)

        # prepare resampled walkers for running new state changes
        walkers = resampled_walkers

        # run the simulation monitor to get metrics on everything
        if self.monitor is not None:
            logging.info("Running monitoring")
            self.monitor.cycle_monitor(self, walkers)


        # we also return a list of the "filters" which are the
        # classes that are run on the initial walkers to produce
        # the final walkers. THis is to satisfy a future looking
        # interface in which the order and components of these
        # filters are completely parametrizable. This may or may
        # not be implemented in a future release of wepy but this
        # interface is assumed by the orchestration classes for
        # making snapshots of the simulations. The receiver of
        # these should perform the copy to make sure they aren't
        # mutated. We don't do this here for efficiency.
        filters = [self.runner, self.boundary_conditions, self.resampler]

        logging.info("Done: returning walkers")
        return walkers, filters

    @log_call
    def init(self,
             num_workers=None,
             continue_run=None,
    ):
        """Initialize wepy configuration components for use at runtime.

        This `init` method is different than the constructor
        `__init__` method and instead calls the special `init` method
        on all wepy components (runner, resampler, boundary
        conditions, and reporters) at runtime.

        This allows for a things that need to be done at runtime before a
        simulation begins, e.g. opening files, that you don't want done at
        construction time.

        It calls the `init` methods on:

        - work_mapper
        - reporters

        Passes the segment_func of the runner and the number of
        workers to the work_mapper.

        Passes the following things to each reporter `init` method:

        - init_walkers
        - runner
        - resampler
        - boundary_conditions
        - work_mapper
        - reporters
        - continue_run

        Parameters
        ----------
        num_workers : int
            The number of workers to use in the work mapper.
             (Default value = None)
        continue_run : int
            Index of a run this one is continuing.
             (Default value = None)

        """


        logging.info("Starting simulation")

        # initialize the monitoring object

        # TODO: do we need to supply the port here? I don't want to
        # add it to the interface... Should be pre-parametrized
        if self.monitor is not None:
            self.monitor.init()

        # initialize the work_mapper with the function it will be
        # mapping and the number of workers, this may include things like starting processes
        # etc.
        self.work_mapper.init(
            segment_func=self.runner.run_segment,
            num_workers=num_workers,
        )


        # init the reporter
        for reporter in self.reporters:
            reporter.init(init_walkers=self.init_walkers,
                          runner=self.runner,
                          resampler=self.resampler,
                          boundary_conditions=self.boundary_conditions,
                          work_mapper=self.work_mapper,
                          reporters=self.reporters,
                          continue_run=continue_run)

    def cleanup(self):
        """Perform cleanup actions for wepy configuration components.

        Allow components to perform actions before ending the main
        simulation manager process.

        Calls the `cleanup` method on:

        - work_mapper
        - reporters

        Passes nothing to the work mapper.

        Passes the following to each reporter:

        - runner
        - work_mapper
        - resampler
        - boundary_conditions
        - reporters

        """

        if self.monitor is not None:
            self.monitor.cleanup()

        # cleanup the mapper
        self.work_mapper.cleanup()

        # cleanup things associated with the reporter
        for reporter in self.reporters:
            reporter.cleanup(runner=self.runner,
                             work_mapper=self.work_mapper,
                             resampler=self.resampler,
                             boundary_conditions=self.boundary_conditions,
                             reporters=self.reporters)


    def run_simulation_by_time(self, run_time, segments_length, num_workers=None):
        """Run a simulation for a certain amount of time.

        This starts timing as soon as this is called. If the time
        before running a new cycle is greater than the runtime the run
        will exit after cleaning up. Once a cycle is started it may
        also run over the wall time.

        All this does is provide a run idx to the reporters, which is
        the run that is intended to be continued. This simulation
        manager knows no details and is left up to the reporters to
        handle this appropriately.

        Parameters
        ----------
        run_time : float
            The time to run in seconds.

        segments_length : int
            The number of steps for each runner segment.

        num_workers : int
            The number of workers to use for the work mapper.
             (Default value = None)

        Returns
        -------
        new_walkers : list of walkers
            The resulting walkers of the cycle

        sim_components : list
            Deep copies of the runner, resampler, and boundary
            conditions objects at the end of the cycle.

        See Also
        --------
        wepy.orchestration.orchestrator.Orchestrator : for running simulations with
            checkpointing, restarting, reporter localization, and configuration hotswapping
            with command line interface.

        """
        start_time = time.time()
        self.init(num_workers=num_workers)
        cycle_idx = 0
        walkers = self.init_walkers
        while time.time() - start_time < run_time:

            logging.info("starting cycle {} at time {}".format(cycle_idx, time.time() - start_time))

            walkers, filters = self.run_cycle(walkers, segments_length, cycle_idx)

            logging.info("ending cycle {} at time {}".format(cycle_idx, time.time() - start_time))


            cycle_idx += 1

        logging.info("Cleaning up simulation")
        self.cleanup()

        return walkers, deepcopy(filters)

    @log_call(
        include_args=[
            'n_cycles',
            'segment_lengths',
            'num_workers',
        ],
        include_result=False)
    def run_simulation(self, n_cycles,
                       segment_lengths,
                       num_workers=None,
    ):
        """Run a simulation for an explicit number of cycles.

        Parameters
        ----------
        n_cycles : int
            Number of cycles to perform.

        segment_lengths : int
            The number of steps for each runner segment.

        num_workers : int
            The number of workers to use for the work mapper.
             (Default value = None)


        Returns
        -------
        new_walkers : list of walkers
            The resulting walkers of the cycle

        sim_components : list
            Deep copies of the runner, boundary conditions, and
            resampler objects at the end of the simulation.

        See Also
        --------
        wepy.orchestration.orchestrator.Orchestrator : for running simulations with
            checkpointing, restarting, reporter localization, and configuration hotswapping
            with command line interface.

        """

        self.init(num_workers=num_workers)

        if type(segment_lengths) == int:
            segment_lengths = [segment_lengths for _ in range(n_cycles)]

        walkers = self.init_walkers

        # the main cycle loop
        with start_action(action_type="Simulation Loop") as simloop_cx:
            for cycle_idx in range(n_cycles):

                walkers, filters = self.run_cycle(walkers, segment_lengths[cycle_idx], cycle_idx)

                # run the simulation monitor to get metrics on everything
                if self.monitor is not None:
                    self.monitor.cycle_monitor(self, walkers)

        self.cleanup()

        return walkers, deepcopy(filters)

    def continue_run_simulation(self,
                                run_idx,
                                n_cycles,
                                segment_lengths,
                                num_workers=None,
    ):
        """Continue a simulation. All this does is provide a run idx to the
        reporters, which is the run that is intended to be
        continued. This simulation manager knows no details and is
        left up to the reporters to handle this appropriately.

        Parameters
        ----------
        run_idx : int
            Index of the run you are continuing.

        n_cycles : int
            Number of cycles to perform.

        segment_lengths : int
            The number of steps for each runner segment.

        num_workers : int
            The number of workers to use for the work mapper.
             (Default value = None)

        Returns
        -------
        new_walkers : list of walkers
            The resulting walkers of the cycle

        sim_components : list
            Deep copies of the runner, resampler, and boundary
            conditions objects at the end of the cycle.

        See Also
        --------
        wepy.orchestration.orchestrator.Orchestrator : for running simulations with
            checkpointing, restarting, reporter localization, and configuration hotswapping
            with command line interface.

        """

        self.init(num_workers=num_workers,
                  continue_run=run_idx)

        walkers = self.init_walkers
        # the main cycle loop
        for cycle_idx in range(n_cycles):
            walkers, filters = self.run_cycle(walkers, segment_lengths[cycle_idx], cycle_idx)

            # run the simulation monitor to get metrics on everything
            if self.monitor is not None:
                self.monitor.cycle_monitor(self, walkers)


        self.cleanup()

        return walkers, filters


    def continue_run_simulation_by_time(self, run_idx, run_time, segments_length, num_workers=None):
        """Continue a simulation with a separate run by time.

        This starts timing as soon as this is called. If the time
        before running a new cycle is greater than the runtime the run
        will exit after cleaning up. Once a cycle is started it may
        also run over the wall time.

        All this does is provide a run idx to the reporters, which is
        the run that is intended to be continued. This simulation
        manager knows no details and is left up to the reporters to
        handle this appropriately.

        Parameters
        ----------
        run_idx : int
            Deep copies of the runner, resampler, and boundary
            conditions objects at the end of the cycle.


        See Also
        --------
        wepy.orchestration.orchestrator.Orchestrator : for running simulations with
            checkpointing, restarting, reporter localization, and configuration hotswapping
            with command line interface.

        """

        start_time = time.time()

        self.init(num_workers=num_workers,
                  continue_run=run_idx)

        cycle_idx = 0
        walkers = self.init_walkers
        while time.time() - start_time < run_time:

            logging.info("starting cycle {} at time {}".format(cycle_idx, time.time() - start_time))

            walkers, filters = self.run_cycle(walkers, segments_length, cycle_idx)

            logging.info("ending cycle {} at time {}".format(cycle_idx, time.time() - start_time))

            # run the simulation monitor to get metrics on everything
            if self.monitor is not None:
                self.monitor.cycle_monitor(self, walkers)

            cycle_idx += 1

        self.cleanup()

        return walkers, filters
